"""Initialize seach result."""

from .mascot import Mascot
from .maxquant import MaxQuant
from .msamanda import MSAmanda
from .msfragger import MSFragger
from .sage import Sage
from .scout import Scout
from .xisearch import Xisearch
